from unittest import TestCase

from pythautomata.base_types.sequence import Sequence
from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines
from pythautomata.base_types.symbol import SymbolStr



class TestMooreMachinesLastSymbolQueries(TestCase):

    def test_empty_sequence(self):
        moore_machine = SampleMooreMachines.get_3_states_automaton()
        emptySequence = Sequence()

        res = moore_machine.last_symbol(emptySequence)

        expected_result = moore_machine.initial_state.value

        return self.assertEqual(expected_result, res)

    def test_non_empty_sequence(self):
        moore_machine = SampleMooreMachines.get_3_states_automaton()
        a_a_symbols = [SymbolStr(SampleMooreMachines.get_a_symbol()), 
                        SymbolStr(SampleMooreMachines.get_a_symbol())]
        non_empty_sequence = Sequence(a_a_symbols)

        res = moore_machine.last_symbol(non_empty_sequence)

        expected_result = SymbolStr(SampleMooreMachines.get_c_symbol())

        return self.assertEqual(expected_result, res)



from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.guard import Guard


class IntersectionGuard(Guard):
    def __init__(self, *guards: Guard):
        self.guards = list(guards)

    def matches(self, s: Symbol):
        # TODO add concurrency
        return all(map(lambda g: g.matches(s), self.guards))

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (u" \u2229 ").join(map(str, self.guards))



from typing import Union

from pythautomata.abstract.finite_automaton import FiniteAutomataComparator
from pythautomata.abstract.finite_automaton import FiniteAutomaton as FA
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.state import State
from pythautomata.base_types.symbol import Symbol
from pythautomata.base_types.symbolic_state import SymbolicState


class HopcroftKarpComparisonStrategy(FiniteAutomataComparator):
    def are_equivalent(self, automaton1: FA, automaton2: FA) -> bool:
        try:
            return self._inner_are_equivalent(automaton1, automaton2) is None
        except ValueError:
            return False

    def get_counterexample_between(self, automaton1: FA, automaton2: FA) -> Union[Sequence, None]:
        counterexample = self._inner_are_equivalent(automaton1, automaton2)
        return counterexample

    def equivalent_output(self, observation1, observation2) -> bool:
        return observation1 == observation2

    def _inner_are_equivalent(self, fa1: FA, fa2: FA) -> Union[Sequence, None]:
        if fa1.has_full_alphabet and fa2.has_full_alphabet:
            if not fa1.alphabet == fa2.alphabet:
                raise ValueError('Alphabets are not equivalent.')

        # symbols is the union of both finite automata's alphabets
        # because one or both of the automata might not have a full alphabet
        symbols = list(fa1.alphabet.symbols | fa2.alphabet.symbols)
        aut1_new_transitions = self._generate_initial_table(fa1, symbols)
        aut2_new_transitions = self._generate_initial_table(fa2, symbols)
        should_be_equivalent_states = {(
            fa1.initial_states, fa2.initial_states): Sequence()}

        checked_equivalence = dict()

        aut1_completed = False
        aut2_completed = False

        while len(should_be_equivalent_states) > 0 or not (aut1_completed and aut2_completed):

            for state_pair, sequence in should_be_equivalent_states.copy().items():
                (aut1_states, aut2_states) = state_pair

                checked_equivalence[frozenset(
                    aut1_states)] = frozenset(aut2_states)

                if not frozenset(aut1_states) in aut1_new_transitions:
                    aut1_next_states = [None for _ in symbols]
                else:
                    aut1_next_states = aut1_new_transitions[frozenset(
                        aut1_states)]

                if not frozenset(aut2_states) in aut2_new_transitions:
                    aut2_next_states = [None for _ in symbols]
                else:
                    aut2_next_states = aut2_new_transitions[frozenset(
                        aut2_states)]

                for sym_index, symbol in enumerate(symbols):
                    aut1_next_states_for_sym = aut1_next_states[sym_index]
                    aut2_next_states_for_sym = aut2_next_states[sym_index]

                    if aut1_next_states_for_sym is None:
                        aut1_next_states_for_sym = self._fill_transitions_for(
                            list(aut1_states), symbol, fa1.hole)

                    if aut2_next_states_for_sym is None:
                        aut2_next_states_for_sym = self._fill_transitions_for(
                            list(aut2_states), symbol, fa2.hole)

                    aut1_reaches_final = self._reaches_final_state(
                        aut1_next_states_for_sym)
                    aut2_reaches_final = self._reaches_final_state(
                        aut2_next_states_for_sym)
                    if aut1_reaches_final != aut2_reaches_final:
                        return sequence + symbol

                    aut1_next_states[sym_index] = aut1_next_states_for_sym
                    aut2_next_states[sym_index] = aut2_next_states_for_sym

                    if self._has_key_or_diff_content(aut1_next_states_for_sym, checked_equivalence, aut2_next_states_for_sym):
                        should_be_equivalent_states[(frozenset(aut1_next_states_for_sym), frozenset(
                            aut2_next_states_for_sym))] = sequence + symbol

                if frozenset(aut1_states) in checked_equivalence:
                    if (aut1_states, checked_equivalence[frozenset(aut1_states)]) in should_be_equivalent_states:
                        del should_be_equivalent_states[(
                            aut1_states, checked_equivalence[frozenset(aut1_states)])]

                aut1_new_transitions[frozenset(aut1_states)] = aut1_next_states
                aut2_new_transitions[frozenset(aut2_states)] = aut2_next_states

            aut1_completed = self._check_table_is_completed(
                aut1_new_transitions)
            aut2_completed = self._check_table_is_completed(
                aut2_new_transitions)
        return None

    def _has_key_or_diff_content(self, aut1_next_states_for_sym, checked_equivalence, aut2_next_states_for_sym):
        contains_key_or_has_different_content = frozenset(aut1_next_states_for_sym) not in checked_equivalence or \
            (checked_equivalence[frozenset(aut1_next_states_for_sym)] != frozenset(
                aut2_next_states_for_sym))
        return contains_key_or_has_different_content

    def _check_table_is_completed(self, new_transitions):
        completed = True
        for next_states in new_transitions.values():
            for next_state_for_symb in next_states:
                if len(next_state_for_symb) > 0 and not frozenset(next_state_for_symb) in new_transitions.keys():
                    completed = False
                    break
        return completed

    def _reaches_final_state(self, next_states_for_sym):
        return any(True for state in next_states_for_sym if state.is_final)

    def _generate_initial_table(self, automaton, symbols: list[Symbol]) -> dict[frozenset[list[Union[State, SymbolicState]]], list[Union[State, SymbolicState]]]:
        initial_states = list(automaton.initial_states)
        next_states_after_initial = self._get_next_states_from_state(
            automaton, initial_states, symbols)
        new_transitions = {
            frozenset(initial_states): next_states_after_initial}
        return new_transitions

    def _get_next_states_from_state(self, fa, states: list[Union[State, SymbolicState]], symbols: list[Symbol]) -> list[list[Union[State, SymbolicState]]]:
        aut_result = list(map(lambda symbol: self._fill_transitions_for(
            states, symbol, fa.hole), symbols))
        return aut_result

    def _fill_transitions_for(self, states: list[Union[State, SymbolicState]], symbol, hole):
        result: list[Union[State, SymbolicState]] = []
        for state in states:
            next_states = state.next_states_for(symbol)
            if not hole in next_states:
                result.extend(next_state for next_state in next_states
                              if next_state not in result)
        return [hole] if len(result) == 0 else result



from graphviz import Digraph
from pythautomata.abstract.pdfa_model_exporting_strategy import PDFAModelExportingStrategy


class WFAImageExporter(PDFAModelExportingStrategy):

    def export(self, model, path=None):
        graph = Digraph('weighted_automaton', format='png')
        graph.attr(rankdir='LR', margin='0', size='15')

        graph.attr('node', shape='circle')
        states = sorted(model.weighted_states, key=lambda x: x.name)
        for state in states:
            label = str(state.final_weight)
            if state.initial_weight == 1:
                graph.node(state.name, label, shape='diamond')
            else:
                graph.node(state.name, label)
            transitions = dict()
            for symbol, weighted_transitions in state.transitions_list.items():
                for weighted_transition in weighted_transitions:
                    if (state.name, weighted_transition[0].name) in transitions.keys():
                        new = transitions[(state.name, weighted_transition[0].name)] + "\n" + str(symbol) + "-" + \
                              str(weighted_transition[1])
                    else:
                        new = str(symbol) + "-" + str(weighted_transition[1])
                    transitions[(state.name, weighted_transition[0].name)] = new
            for key in transitions.keys():
                s_from, s_to = key
                graph.edge(s_from, s_to, str(transitions[key]))

        graph.render(self.get_path_for(path, model), cleanup=True)



from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.utilities.pdfa_operations import get_representative_sample


class TestPDFAOperations(TestCase):

    def test_1(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        sample = get_representative_sample(pdfa, 10)
        self.assertTrue(True)

    def test_2(self):
        pdfa = WeightedTomitasGrammars.get_automaton_2()
        sample = get_representative_sample(pdfa, 20)
        self.assertTrue(True)

    def test_3(self):
        pdfa = WeightedTomitasGrammars.get_automaton_3()
        sample = get_representative_sample(pdfa, 30)
        self.assertTrue(True)

    def test_4(self):
        pdfa = WeightedTomitasGrammars.get_automaton_4()
        sample = get_representative_sample(pdfa, 400)
        self.assertTrue(True)

    def test_5(self):
        pdfa = WeightedTomitasGrammars.get_automaton_5()
        sample = get_representative_sample(pdfa, 0)
        self.assertTrue(True)

    def test_6(self):
        pdfa = WeightedTomitasGrammars.get_automaton_6()
        sample = get_representative_sample(pdfa, 1)
        self.assertTrue(True)

    def test_7(self):
        pdfa = WeightedTomitasGrammars.get_automaton_7()
        self.assertRaises(AssertionError, get_representative_sample, pdfa, -1)



from abc import ABC, abstractmethod
from pythautomata.base_types.symbol import Symbol

class Guard(ABC):

    
    @abstractmethod
    def matches(self, s:Symbol) -> bool: 
        """checks if the symbol s matches the guard

        Args:
            s (Symbol): symbol to check

        Returns:
            bool: True if the symbol matches the guard, False otherwise
        """        
        pass



from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.model_comparators.wfa_tolerance_comparison_strategy import WFAToleranceComparator

binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
zero = binaryAlphabet['0']
one = binaryAlphabet['1']


class WeightedTomitasGrammars:
    """
    Class containing weighted (probabilistic) versions of automata from paper:    
        Tomita, M. (1982). 
        Dynamic Construction of Finite Automata from examples using Hill-climbing. 
        Proceedings of the Fourth Annual Conference of the Cognitive Science Society 
        (p./pp. 105--108), Ann Arbor, Michigan.

    The weighted version is taken from paper:
        Weiss, Gail & Goldberg, Yoav & Yahav, Eran. (2019). 
        Learning Deterministic Weighted Automata with Queries and Counterexamples. 

    Methods
    -------   
    get_all_automata: list(ProbabilisticDeterministicFiniteAutomaton)
        returns a list containing all weigthed automata defined in this class

    get_automaton_1: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 1 from the paper

    get_automaton_2: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 2 from the paper

    get_automaton_3: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 3 from the paper

    get_automaton_4: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 4 from the paper

    get_automaton_5: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 5 from the paper

    get_automaton_6: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 6 from the paper

    get_automaton_7: ProbabilisticDeterministicFiniteAutomaton
        returns the weighted automaton 7 from the paper
    """

    @staticmethod
    def get_all_automata():
        """
        Method returning a list of all automata of the class

        Returns
        -------
        list(ProbabilisticDeterministicFiniteAutomaton)
            all automata defined in the class
        """
        return [
            WeightedTomitasGrammars.get_automaton_1(),
            WeightedTomitasGrammars.get_automaton_2(),
            WeightedTomitasGrammars.get_automaton_3(),
            WeightedTomitasGrammars.get_automaton_4(),
            WeightedTomitasGrammars.get_automaton_5(),
            WeightedTomitasGrammars.get_automaton_6(),
            WeightedTomitasGrammars.get_automaton_7()
        ]

    @staticmethod
    def get_automaton_1():
        """
        method with specification of automaton 1 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 1
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)

        q0.add_transition(zero, q1, 0.665)
        q0.add_transition(one, q0, 0.285)
        q1.add_transition(zero, q1, 0.285)
        q1.add_transition(one, q1, 0.665)

        states = {q0, q1}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas1")

    @staticmethod
    def get_automaton_2():
        """
        method with specification of automaton 2 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 2
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)

        q0.add_transition(zero, q2, 0.665)
        q0.add_transition(one, q1, 0.285)
        q1.add_transition(zero, q0, 0.285)
        q1.add_transition(one, q2, 0.665)
        q2.add_transition(zero, q2, 0.285)
        q2.add_transition(one, q2, 0.665)

        states = {q0, q1, q2}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas2")

    @staticmethod
    def get_automaton_3():
        """
        method with specification of automaton 3 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 3
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)
        q3 = WeightedState("q3", 0, 0.05)
        q4 = WeightedState("q4", 0, 0.05)

        q0.add_transition(zero, q0, 0.665)
        q0.add_transition(one, q1, 0.285)
        q1.add_transition(zero, q2, 0.665)
        q1.add_transition(one, q0, 0.285)
        q2.add_transition(zero, q3, 0.285)
        q2.add_transition(one, q4, 0.665)
        q3.add_transition(zero, q2, 0.665)
        q3.add_transition(one, q3, 0.285)
        q4.add_transition(zero, q4, 0.285)
        q4.add_transition(one, q4, 0.665)

        states = {q0, q1, q2, q3, q4}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas3")

    @staticmethod
    def get_automaton_4():
        """
        method with specification of automaton 4 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 4
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)
        q3 = WeightedState("q3", 0, 0.05)

        q0.add_transition(zero, q1, 0.665)
        q0.add_transition(one, q0, 0.285)
        q1.add_transition(zero, q2, 0.665)
        q1.add_transition(one, q0, 0.285)
        q2.add_transition(zero, q3, 0.665)
        q2.add_transition(one, q0, 0.285)
        q3.add_transition(zero, q3, 0.285)
        q3.add_transition(one, q3, 0.665)

        states = {q0, q1, q2, q3}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas4")

    @staticmethod
    def get_automaton_5():
        """
        method with specification of automaton 5 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 5
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)
        q3 = WeightedState("q3", 0, 0.05)

        q0.add_transition(zero, q3, 0.665)
        q0.add_transition(one, q1, 0.285)
        q1.add_transition(zero, q2, 0.285)
        q1.add_transition(one, q0, 0.665)
        q2.add_transition(zero, q1, 0.285)
        q2.add_transition(one, q3, 0.665)
        q3.add_transition(zero, q0, 0.285)
        q3.add_transition(one, q2, 0.665)

        states = {q0, q1, q2, q3}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas5")

    @staticmethod
    def get_automaton_6():
        """
        method with specification of automaton 6 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 6
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)

        q0.add_transition(zero, q2, 0.665)
        q0.add_transition(one, q1, 0.285)
        q1.add_transition(zero, q0, 0.285)
        q1.add_transition(one, q2, 0.665)
        q2.add_transition(zero, q1, 0.285)
        q2.add_transition(one, q0, 0.665)

        states = {q0, q1, q2}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas6")

    @staticmethod
    def get_automaton_7():
        """
        method with specification of automaton 7 from the paper       

        Returns
        -------
        ProbabilisticDeterministicFiniteAutomaton
            weighted tomita grammar 7
        """
        q0 = WeightedState("q0", 1, 0.05)
        q1 = WeightedState("q1", 0, 0.05)
        q2 = WeightedState("q2", 0, 0.05)
        q3 = WeightedState("q3", 0, 0.05)
        q4 = WeightedState("q4", 0, 0.05)

        q0.add_transition(zero, q0, 0.665)
        q0.add_transition(one, q1, 0.285)
        q1.add_transition(zero, q2, 0.665)
        q1.add_transition(one, q1, 0.285)
        q2.add_transition(zero, q2, 0.665)
        q2.add_transition(one, q3, 0.285)
        q3.add_transition(zero, q4, 0.665)
        q3.add_transition(one, q3, 0.285)
        q4.add_transition(zero, q4, 0.285)
        q4.add_transition(one, q4, 0.665)

        states = {q0, q1, q2, q3, q4}
        comparator = WFAToleranceComparator()
        return ProbabilisticDeterministicFiniteAutomaton(binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas7")



import numpy as np


def are_within_tolerance_limit(obs1, obs2, tolerance):
    assert(len(obs1) == len(obs2))
    return np.all((abs(np.array(obs1) - np.array(obs2)) <= tolerance))


def get_partition(value, partitions):
    assert(value >= 0 and value <= 1)
    limits = np.linspace(0, 1, partitions+1)
    if value == 1:
        return partitions-1
    positions = list(range(len(limits)-1))
    mid_element = int(len(limits)/2)
    while len(positions) > 1:
        if value >= limits[mid_element]:
            positions = positions[int(len(positions)/2):]
        else:
            positions = positions[:int(len(positions)/2)]
        mid_element = positions[int(len(positions)/2)]
    assert(len(positions) == 1)
    return positions[0]


def get_partitions(observation, partitions):
    return np.fromiter((get_partition(xi, partitions) for xi in observation), dtype=int)


def are_in_same_partition(obs1, obs2, partitions):
    assert(len(obs1) == len(obs2))
    return are_same_partition(get_partitions(obs1, partitions),
                              get_partitions(obs2, partitions))


def are_same_partition(partition1, partition2):
    assert(len(partition1) == len(partition2))
    return np.all((abs(partition1 - partition2) == 0))



from typing import Optional
import random

from pythautomata.automata.moore_machine_automaton import MooreMachineAutomaton as MooreMachine
from pythautomata.base_types.sequence import Sequence
from random import seed
from random import randint

class RandomWalkMMComparisonStrategy:
    def __init__(self, number_steps: int, reset_probability: float, random_seed: int = 21) -> None:
        super().__init__()
        self._number_steps = number_steps
        self._reset_probability = reset_probability
        self._seed = random_seed
        seed(self._seed)

    def are_equivalent(self, mm1: MooreMachine, mm2: MooreMachine) -> bool:
        return self.get_counterexample_between(mm1, mm2) is None

    def get_counterexample_between(self, mm1, mm2) -> Optional[Sequence]:
        if mm1._input_alphabet != mm2._input_alphabet:
            raise ValueError("Alphabets are not equivalent.")

        steps = 0
        counter_example = Sequence()
        mm1.reset()
        mm2.reset()
        equivalent_output = self.equivalent_output(
            mm1.last_symbol(counter_example), mm2.last_symbol(counter_example))

        symbols = list(mm1._input_alphabet.symbols)
        symbols.sort()

        while equivalent_output and steps < self._number_steps:
            if random.random() <= self._reset_probability:
                counter_example = Sequence()
                mm1.reset()
                mm2.reset()
            pos = randint(0, len(symbols) - 1)
            random_symbol = symbols[pos]
            counter_example = counter_example.append(random_symbol)
            steps += 1
            equivalent_output = self.equivalent_output(
                mm1.step(random_symbol), mm2.step(random_symbol))
            if not equivalent_output:
                return counter_example
        return None

    def equivalent_output(self, observation1, observation2) -> bool:
        return observation1 == observation2



import random

from pythautomata.automata.deterministic_finite_automaton import DeterministicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.abstract.finite_automaton import FiniteAutomataComparator
from pythautomata.model_comparators.wfa_tolerance_comparison_strategy import WFAToleranceComparator
from typing import Any


def pdfa_from_dfa(dfa: DeterministicFiniteAutomaton, comparator: FiniteAutomataComparator = WFAToleranceComparator(),
                  distributions: int = None, max_shift: float = 0) -> ProbabilisticDeterministicFiniteAutomaton:
    """
    Function that transforms a DFA to a PDFA with random probability distributions for each state.
    Distributions is a predefined integer that defines the number of different distributions states may have (infinite if None)
    And shift is a float definig the ammount of random noise that may be added to such predefined distributions (only used when distributions != None).

    Parameters
    ----------
    dfa : DeterministicFiniteAutomaton
    comparator : FiniteAutomataComparator
    distributions : int
    max_shift: float

    Returns
    -------
    ProbabilisticDeterministicFiniteAutomaton

    """
    alphabet_length = len(dfa.alphabet)
    wfa_states_dict = {state.name: __dfa_state_to_pdfa_state(state.name, state.name == dfa.initial_state.name) for state
                       in dfa.states}
    if distributions is None:
        for state in dfa.states:
            __add_transitions(alphabet_length, state, wfa_states_dict)
    else:
        pool = __generate_pool_of_distributions(alphabet_length, distributions)
        for state in dfa.states:
            distribution = random.choice(pool)
            __add_shift(distribution, max_shift)
            __add_transitions(alphabet_length, state,
                              wfa_states_dict, probability_vector=distribution)

    terminal_symbol = SymbolStr('$')
    return ProbabilisticDeterministicFiniteAutomaton(dfa.alphabet, set(wfa_states_dict.values()), terminal_symbol,
                                                     comparator)


def __generate_pool_of_distributions(alphabet_length, distributions):
    result = []
    for i in range(distributions):
        dist = __generate_random_probability_vector(alphabet_length)
        result.append(dist)
    return result


def __add_shift(distribution, max_shift):
    shift = random.triangular(0, max_shift)
    sign = random.choice([-1, 1])
    total_shift = 0
    for i, elem in enumerate(distribution):
        if i < len(distribution)-1:
            total_shift += shift*sign
            distribution[i] = round(elem + shift*sign, 5)
            sign = sign*-1
            shift = random.triangular(0, max_shift)
        else:
            distribution[i] = round(elem + -1*total_shift, 5)


def __dfa_state_to_pdfa_state(name, initial):
    initial_prob = 1 if initial else 0
    final_prob = None
    wfa_state = WeightedState(name, initial_prob, final_prob)
    return wfa_state


def __generate_random_probability_vector(alphabet_length):
    final_prob = __get_prob(0)
    total_prob = final_prob
    probs = []
    for i in range(alphabet_length):
        prob = round(1.0 - total_prob, 5) if i == alphabet_length - \
            1 else __get_prob(total_prob)
        probs.append(prob)
        total_prob += prob
    probs.append(final_prob)
    return probs


def __add_transitions(alphabet_length, state, wfa_states_dict: dict[Any, WeightedState], probability_vector=None):
    wfa_state = wfa_states_dict[state.name]
    if probability_vector is None:
        probs = __generate_random_probability_vector(alphabet_length)
    else:
        probs = probability_vector
    for i, (symbol, next_state) in enumerate(state.transitions.items()):
        next_state = next(iter(next_state))
        wfa_state.add_transition(
            symbol, wfa_states_dict[next_state.name], probs[i])
    wfa_state.final_weight = probs[-1]


def __get_prob(total_prob):
    prob = random.triangular(0.0, 1.0 - total_prob)
    return round(prob, 5)



from unittest import TestLoader, TestSuite, TextTestRunner

from pythautomata.tests.test_minimizer import TestMinimizer
from pythautomata.tests.test_automata_convertor import TestAutomataConvertor
from pythautomata.tests.test_moore_machines_last_symbol_queries import TestMooreMachinesLastSymbolQueries
from pythautomata.tests.test_pdfa_generator import TestPDFAGenerator
from pythautomata.tests.test_pdfa_serialization import TestPDFASerialization
from pythautomata.tests.test_dfa_serialization import TestDFASerialization
from pythautomata.tests.test_DFA_generators import TestDFAGenerators
from pythautomata.tests.test_automata_comparison import TestAutomataComparison
from pythautomata.tests.test_sequence import TestSequence
from pythautomata.tests.test_dfa_operations import TestDFAOperations
from pythautomata.tests.test_automata_definitions import TestAutomataDefinitions
from pythautomata.tests.test_regex_generator import TestRegexGenerator
from pythautomata.tests.test_pdfa_metrics import TestPDFAMetrics
from pythautomata.tests.test_simple_DFA_generator import TestSimpleDFAGenerator
from pythautomata.tests.test_pdfa_operations import TestPDFAOperations
from pythautomata.tests.test_pdfa_last_token_queries import TestPDFALastTokenQueries
from pythautomata.tests.test_dfa_to_moore_machine import TestDFAToMooreMachine
from pythautomata.tests.test_dfa_exporting import TestDFAExporting
from pythautomata.tests.test_dfa_loading import TestDFALoading
from pythautomata.tests.test_mm_exporting import TestMMExporting
from pythautomata.tests.test_mm_loading import TestMMLoading


def get_all_tests():
    return [TestDFALoading, TestDFAExporting, TestSimpleDFAGenerator, TestMinimizer,
            TestAutomataConvertor, TestDFAGenerators, TestAutomataComparison, TestSequence, 
            TestDFAOperations, TestAutomataDefinitions, TestPDFAGenerator, TestRegexGenerator, 
            TestPDFASerialization,TestDFASerialization, TestPDFAMetrics, TestPDFAOperations, 
            TestPDFALastTokenQueries, TestMooreMachinesLastSymbolQueries, 
            TestDFAToMooreMachine, TestMMLoading, TestMMExporting]



def run():
    test_classes_to_run = get_all_tests()
    loader = TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    meta_suite = TestSuite(suites_list)
    TextTestRunner().run(meta_suite)


if __name__ == '__main__':
    run()



import pickle
from unittest import TestCase

from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars


class TestPDFASerialization(TestCase):

    def test_1(self):
        pdfa = WeightedTomitasGrammars.get_automaton_1()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_2(self):
        pdfa = WeightedTomitasGrammars.get_automaton_2()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_3(self):
        pdfa = WeightedTomitasGrammars.get_automaton_3()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_4(self):
        pdfa = WeightedTomitasGrammars.get_automaton_4()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_5(self):
        pdfa = WeightedTomitasGrammars.get_automaton_5()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_6(self):
        pdfa = WeightedTomitasGrammars.get_automaton_6()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)

    def test_7(self):
        pdfa = WeightedTomitasGrammars.get_automaton_7()
        data = pickle.dumps(pdfa)
        pdfa2 = pickle.loads(data)
        self.assertEqual(pdfa, pdfa2)



from genericpath import isdir
from os import makedirs
from pathlib import Path
from pythautomata.model_exporters.image_exporting_mm_strategy import ImageExportingMMStrategy


class DotExportingMMStrategy():
    def export(self, model, path=None):
        graph = ImageExportingMMStrategy().create_graph(model)
        path = self.get_path_for(path, model)
        path = str(path) + '.dot'

        dot_code = graph.source

        with open(path, "w+", encoding="utf-8") as f:
            f.write(dot_code)

    def get_path_for(self, path: str, model):
        if path is None:
            name = model._name
            path = "output_models/" + \
                ("" if name is None else f"{name}")
        if not isdir(path):
            makedirs(path)
        return Path(path, model._name)



from pythautomata.abstract.boolean_model import BooleanModel
from pythautomata.abstract.finite_automaton import FiniteAutomataComparator, FiniteAutomaton
from pythautomata.abstract.model_exporting_strategy import ModelExportingStrategy
from pythautomata.base_types.symbolic_state import SymbolicState
from pythautomata.base_types.sequence import Sequence
from pythautomata.base_types.alphabet import Alphabet
import uuid

from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import HopcroftKarpComparisonStrategy


class SymbolicFiniteAutomaton(FiniteAutomaton, BooleanModel):
    """Implementation of a Symbolic Automaton

    Attributes
    ----------
    states: set[State]
        Set containing the SFA's states
    initial_state: State
        Initial state of the SFA. Also included in "states"
    """

    def __init__(self, alphabet: Alphabet, initial_state: SymbolicState, states: set[SymbolicState], name: str = None,
                 exportingStrategies: list[ModelExportingStrategy] = [], hole: SymbolicState = SymbolicState("Hole"),
                 comparator: FiniteAutomataComparator = HopcroftKarpComparisonStrategy()):
        self._name = 'DFA - ' + str(uuid.uuid4().hex) if name is None else name
        self._alphabet = alphabet
        self.initial_state = initial_state
        self.states = states
        self._set_hole(hole)
        self._exporting_strategies = exportingStrategies

    def _set_hole(self, hole: SymbolicState) -> None:
        self._hole = hole
        # hole's hole state is itself
        self._hole.add_hole_transition(self.hole)
        for state in self.states:
            state.add_hole_transition(hole)

    @property
    def hole(self):
        return self._hole

    @property
    def initial_states(self) -> frozenset:
        return frozenset([self.initial_state])

    def accepts(self, sequence: Sequence) -> bool:
        actual_state = self.initial_state
        for symbol in sequence:
            actual_state = actual_state.next_state_for(symbol)
        return actual_state.is_final

    @property
    def has_full_alphabet(self) -> bool:
        return False



from itertools import chain
from unittest import TestCase

from pythautomata.automata.deterministic_finite_automaton import \
    DeterministicFiniteAutomaton as DFA
from pythautomata.automata.symbolic_finite_automaton import SymbolicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
    ProbabilisticDeterministicFiniteAutomaton
from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
from pythautomata.automata_definitions.bollig_habermehl_kern_leucker_automata import \
    BolligHabermehlKernLeuckerAutomata
from pythautomata.automata_definitions.omlin_giles_automata import \
    OmlinGilesAutomata
from pythautomata.automata_definitions.other_automata import OtherAutomata
from pythautomata.automata_definitions.tomitas_grammars import TomitasGrammars
from pythautomata.automata_definitions.weighted_tomitas_grammars import WeightedTomitasGrammars
from pythautomata.base_types.alphabet import Alphabet
from pythautomata.base_types.symbol import SymbolStr
from pythautomata.base_types.symbolic_state import SymbolicState
from pythautomata.model_comparators.dfa_comparison_strategy import \
    DFAComparisonStrategy as DFAComparator
from pythautomata.model_comparators.random_walk_comparison_strategy import \
    RandomWalkComparisonStrategy
from pythautomata.model_comparators.hopcroft_karp_comparison_strategy import \
    HopcroftKarpComparisonStrategy as HopcroftKarpComparison
from pythautomata.model_comparators.random_walk_mm_comparison_strategy import RandomWalkMMComparisonStrategy
from pythautomata.utilities.automata_converter import AutomataConverter
from pythautomata.model_comparators.wfa_tolerance_comparison_strategy import WFAToleranceComparator
from pythautomata.model_comparators.wfa_quantization_comparison_strategy import WFAQuantizationComparator
from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines


class TestAutomataComparison(TestCase):

    def test_equivalence_reflexiveness(self):
        mergedAutomata = list(chain(TomitasGrammars.get_all_automata(),
                                    OmlinGilesAutomata.get_all_automata(),
                                    OtherAutomata.get_all_automata()))
        for automaton in mergedAutomata:
            self.assertAreEquivalentFAs(automaton, automaton)

    def test_equivalence_minimized_automaton_1(self):
        automaton = OtherAutomata.get_automaton_1()
        minimized_automaton = OtherAutomata.get_automaton_1_minimized()
        self.assertAreEquivalentFAs(automaton, minimized_automaton)

    def test_equivalence_minimized_automaton_2(self):
        automaton = OtherAutomata.get_automaton_2()
        minimized_automaton = OtherAutomata.get_automaton_2_minimized()
        self.assertAreEquivalentFAs(automaton, minimized_automaton)

    def test_equivalence_minimized_automaton_3(self):
        automaton = OtherAutomata.get_automaton_3()
        minimized_automaton = OtherAutomata.get_automaton_3_minimized()
        self.assertAreEquivalentFAs(automaton, minimized_automaton)

    def test_equivalence_minimized_automaton_4(self):
        automaton = OtherAutomata.get_automaton_4()
        minimized_automaton = OtherAutomata.get_automaton_4_minimized()
        self.assertAreEquivalentFAs(automaton, minimized_automaton)

    def test_dfa_ndfa_equivalence_1(self):
        dfa = OtherAutomata.get_nfa_1()
        ndfa = OtherAutomata.get_dfa_1()
        self.assertAreEquivalentFAs(dfa, ndfa)

    def test_dfa_ndfa_equivalence_2(self):
        dfa = OtherAutomata.get_nfa_2()
        ndfa = OtherAutomata.get_dfa_2()
        self.assertAreEquivalentFAs(dfa, ndfa)

    def test_dfa_ndfa_equivalence_3(self):
        dfa = OtherAutomata.get_nfa_3()
        ndfa = OtherAutomata.get_dfa_3()
        self.assertAreEquivalentFAs(dfa, ndfa)

    def test_dfa_ndfa_equivalence_4(self):
        dfa = OtherAutomata.get_nfa_4()
        ndfa = OtherAutomata.get_dfa_4()
        self.assertAreEquivalentFAs(dfa, ndfa)

    def test_dfa_ndfa_equivalence_5(self):
        dfa = OtherAutomata.get_nfa_5()
        ndfa = OtherAutomata.get_dfa_5()
        self.assertAreEquivalentFAs(dfa, ndfa)

    def test_e_commerce_equivalent_to_itself(self):
        e_commerce_automata = OtherAutomata.get_reduced_ecommerce_automaton()
        self.assertAreEquivalentFAs(e_commerce_automata, e_commerce_automata)

    def test_e_commerce_equivalence_with_reduced_version(self):
        e_commerce_automata = OtherAutomata.get_reduced_ecommerce_automaton()
        reduced_e_commerce_automata = OtherAutomata.get_reduced_ecommerce_automaton()
        self.assertAreEquivalentFAs(
            e_commerce_automata, reduced_e_commerce_automata)

    def test_comparator_equivalency_when_different_1(self):
        automaton = OtherAutomata.get_nfa_1()
        nonEquivalentDfa = OtherAutomata.get_dfa_2()
        self.assertAreNotEquivalentFAs(automaton, nonEquivalentDfa)

    def test_comparator_equivalency_when_different_2(self):
        automaton = OtherAutomata.get_nfa_2()
        nonEquivalentDfa = OtherAutomata.get_dfa_1()
        self.assertAreNotEquivalentFAs(automaton, nonEquivalentDfa)

    def test_comparator_equivalency_through_counterexample(self):
        automaton = OtherAutomata.get_nfa_1()
        equivalentDfa = OtherAutomata.get_dfa_1()
        comparator = DFAComparator()
        automaton = AutomataConverter.convert_nfa_to_dfa(automaton)
        counterexample = comparator.get_counterexample_between(
            automaton, equivalentDfa)
        self.assertIsNone(counterexample)

    def test_comparator_returns_valid_counterexample(self):
        nfa = OtherAutomata.get_nfa_1()
        nonEquivalentDfa = OtherAutomata.get_dfa_2()
        transformedNfa = AutomataConverter.convert_nfa_to_dfa(nfa)
        comparator = DFAComparator()
        counterexample = comparator.get_counterexample_between(
            transformedNfa, nonEquivalentDfa)
        self.assertIsNotNone(counterexample)
        self.assertTrue(transformedNfa.accepts(counterexample) !=
                        nonEquivalentDfa.accepts(counterexample))

    def test_comparator_returns_valid_counterexample_2(self):
        ecommerce_automaton = OtherAutomata.get_ecommerce_automaton()
        different_ecommerce_automaton = OtherAutomata.get_different_ecommerce_automaton_NFA()
        different_ecommerce_automaton = AutomataConverter.convert_nfa_to_dfa(
            different_ecommerce_automaton)
        comparator = DFAComparator()
        counterexample = comparator.get_counterexample_between(
            ecommerce_automaton, different_ecommerce_automaton)
        self.assertIsNotNone(counterexample)
        self.assertTrue(ecommerce_automaton.accepts(counterexample) !=
                        different_ecommerce_automaton.accepts(counterexample))

    def test_nfa_learning_FAs(self):
        dfa = BolligHabermehlKernLeuckerAutomata.get_first_example_DFA()
        nfa = BolligHabermehlKernLeuckerAutomata.get_first_example_NFA()
        self.assertAreEquivalentFAs(dfa, nfa)
        self.assertAreEquivalentFAs(nfa, dfa)

    def _areEquivalentFAs(self, automaton1, automaton2):
        if type(automaton1) == DFA and type(automaton2) == DFA:
            comparator = DFAComparator()
        else:
            comparator = HopcroftKarpComparison()
        areEquivalent = comparator.are_equivalent(automaton1, automaton2)
        return areEquivalent

    def assertAreEquivalentFAs(self, automaton1, automaton2):
        assert(self._areEquivalentFAs(automaton1, automaton2))

    def assertAreNotEquivalentFAs(self, automaton1, automaton2):
        self.assertFalse(self._areEquivalentFAs(automaton1, automaton2))

    def test_wfa_equivalence_reflexiveness(self):
        weightedTomitasAutomata = WeightedTomitasGrammars.get_all_automata()
        comparator = WFAToleranceComparator()
        for wfa in weightedTomitasAutomata:
            assert(comparator.are_equivalent(wfa, wfa))

    def test_wfa_equivalence_false_case(self):
        weightedTomitasAutomata = WeightedTomitasGrammars.get_all_automata()
        comparator = WFAToleranceComparator()
        for i in range(len(weightedTomitasAutomata)):
            for j in range(len(weightedTomitasAutomata)):
                if i != j:
                    wfa1 = weightedTomitasAutomata[i]
                    wfa2 = weightedTomitasAutomata[j]
                    assert(not comparator.are_equivalent(wfa1, wfa2))

    def test_tomita4_vs_single_state_pdfa(self):
        weightedTomita4 = WeightedTomitasGrammars.get_automaton_4()
        from pythautomata.base_types.symbol import SymbolStr
        from pythautomata.base_types.alphabet import Alphabet
        from pythautomata.automata.wheighted_automaton_definition.weighted_state import WeightedState
        from pythautomata.automata.wheighted_automaton_definition.probabilistic_deterministic_finite_automaton import \
            ProbabilisticDeterministicFiniteAutomaton
        q0 = WeightedState("q0", 1, 0.05)
        binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        zero = binaryAlphabet['0']
        one = binaryAlphabet['1']
        q0.add_transition(zero, q0, 0.665)
        q0.add_transition(one, q0, 0.285)

        states = {q0}
        comparator = WFAToleranceComparator()
        single_state_wfa = ProbabilisticDeterministicFiniteAutomaton(
            binaryAlphabet, states, SymbolStr("$"), comparator, "WeightedTomitas4")
        counterexample = comparator.get_counterexample_between(
            weightedTomita4, single_state_wfa)
        assert(counterexample is not None)

    def test_wfa_quantization_equivalence_reflexiveness(self):
        weightedTomitasAutomata = WeightedTomitasGrammars.get_all_automata()
        comparator = WFAQuantizationComparator(100)
        for wfa in weightedTomitasAutomata:
            assert(comparator.are_equivalent(wfa, wfa))

    def test_wfa_quantization_equivalence_false_case(self):
        weightedTomitasAutomata = WeightedTomitasGrammars.get_all_automata()
        comparator = WFAQuantizationComparator(100)
        for i in range(len(weightedTomitasAutomata)):
            for j in range(len(weightedTomitasAutomata)):
                if i != j:
                    wfa1 = weightedTomitasAutomata[i]
                    wfa2 = weightedTomitasAutomata[j]
                    assert(not comparator.are_equivalent(wfa1, wfa2))

    def test_quantization_equivalence_tomita4_vs_single_state_pdfa(self):
        weightedTomita4 = WeightedTomitasGrammars.get_automaton_4()
        q0 = WeightedState("q0", 1, 0.05)
        binaryAlphabet = Alphabet(frozenset((SymbolStr('0'), SymbolStr('1'))))
        zero = binaryAlphabet['0']
        one = binaryAlphabet['1']
        q0.add_transition(zero, q0, 0.665)
        q0.add_transition(one, q0, 0.285)

        states = {q0}
        comparator = WFAQuantizationComparator(100)
        single_state_wfa = ProbabilisticDeterministicFiniteAutomaton(
            binaryAlphabet, states, SymbolStr("$"), comparator, "Single State PDFA")
        counterexample = comparator.get_counterexample_between(
            weightedTomita4, single_state_wfa)
        assert (counterexample is not None)

    def test_tomita1_vs_empty_sfa(self):
        tomita1 = TomitasGrammars.get_automaton_1()
        comparator = HopcroftKarpComparison()
        init_state = SymbolicState('', is_final=True)
        sfa = SymbolicFiniteAutomaton(
            Alphabet(frozenset()), init_state, {init_state})
        assert(not comparator.are_equivalent(tomita1, sfa))

    def test_equivalence_reflexiveness_random_walk(self):
        mergedAutomata = list(chain(TomitasGrammars.get_all_automata()))
        comparison_strategy1 = RandomWalkComparisonStrategy(1000, 0.01)
        comparison_strategy2 = RandomWalkComparisonStrategy(1000, 0.01)
        for i in range(len(mergedAutomata)):
            for j in range(len(mergedAutomata)):
                if i == j:
                    self.assertTrue(comparison_strategy1.are_equivalent(
                        mergedAutomata[i], mergedAutomata[j].clone()))
                else:
                    self.assertFalse(
                        comparison_strategy2.are_equivalent(mergedAutomata[i], mergedAutomata[j]))

    def test_equivalence_random_walk_for_mm(self):
        mergedAutomata = list(chain(SampleMooreMachines.get_all_automata()))
        mergedAutomata2 = list(chain(SampleMooreMachines.get_all_automata()))
        comparison_strategy = RandomWalkMMComparisonStrategy(1000, 0.01)

        for i in range(len(mergedAutomata)):
            for j in range(len(mergedAutomata2)):
                if i == j:
                    self.assertTrue(comparison_strategy.are_equivalent(
                        mergedAutomata[i], mergedAutomata2[j]))
                else:
                    try:
                        self.assertFalse(
                            comparison_strategy.are_equivalent(mergedAutomata[i], mergedAutomata2[j]))
                    except:
                        pass



from unittest import TestCase

from pythautomata.automata_definitions.sample_moore_machines import SampleMooreMachines
from pythautomata.model_exporters.standard_dot_exporting_mm_strategy import StandardDotMMExportingStrategy
from pythautomata.utilities.standard_dot_notation_mm_importer import StandardDotMMNotationImporter
class TestMMLoading(TestCase):

    def test_1(self):
        for mm in SampleMooreMachines.get_all_automata():
            StandardDotMMExportingStrategy().export(mm, "./output_models/tests")
            path = './output_models/tests'+'/Standard '+mm._name+'.dot'
            loaded_mm = StandardDotMMNotationImporter.import_automata(path)
            self.assertEqual(mm, loaded_mm)